from datetime import datetime, timezone
from typing import Dict
from unittest.mock import call, patch, sentinel
from urllib.parse import urlencode

from django.contrib.auth import get_user_model
from django.test import TestCase, override_settings, tag
from django.urls import reverse
from django.urls.exceptions import NoReverseMatch
from ferda.constants import LoggerResult
from ferda.tests.utils import LoggerMixin
from prequel import (
    PublicRequest,
    PublicRequestId,
    Reference,
    RegistryObjectType,
    RequestOnStatusAction,
    RequestStatus,
)
from prequel.exceptions import CannotBeProcessed, PublicRequestDoesNotExist

from ferda_pubreq.constants import LogEntryType, LogReferenceType

User = get_user_model()


def make_public_request(
    id: int,
    ref_type: RegistryObjectType = RegistryObjectType.DOMAIN,
    ref_value: str = "012-345-6789",
) -> PublicRequest:
    return PublicRequest(
        public_request_id=PublicRequestId("ID " + str(id)),
        numeric_id=id,
        reference=Reference(type=ref_type, value=ref_value),
        type="AuthInfo",
        create_time=datetime(2023, 1, 1, 12, 0, 0, tzinfo=timezone.utc),
        status=RequestStatus.OPEN,
        on_status_action=RequestOnStatusAction.SCHEDULED,
    )


@tag("api")
@override_settings(ROOT_URLCONF="ferda_pubreq.tests.urls")
class GetTest(LoggerMixin, TestCase):
    """Test cases with HTTP GET method (information)."""

    def setUp(self):
        super().setUp()
        user = User.objects.create_superuser("admin")
        self.client.force_login(user)

    # --- bad URL

    @patch("ferda_pubreq.api.PUBREQ_INFO", autospec=True)
    def test_get_bad_url(self, pubreq_client):
        pubreq_client.get_types.return_value = ""
        with self.assertRaises(NoReverseMatch):
            self.client.get(reverse("ferda_pubreq:api:get"))

    # --- get_types

    @patch("ferda_pubreq.api.PUBREQ_INFO", autospec=True)
    def test_get_types(self, pubreq_client):
        get_types_reply: Dict[str, bool] = {
            "AuthInfo": True,
            "LockObject": True,
            "UnlockObject": True,
            "GetDetails": True,
        }
        pubreq_client.get_types.return_value = get_types_reply

        response = self.client.get(reverse("ferda_pubreq:api:get_types"))

        self.assertEqual(pubreq_client.mock_calls, [call.get_types()])
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), get_types_reply)

    # --- get

    @patch("ferda_pubreq.api.PUBREQ_INFO", autospec=True)
    def test_get_ok(self, pubreq_client):
        ref_value: str = "012-345-6789"
        reply = make_public_request(id=1, ref_value=ref_value)
        id = reply.public_request_id
        pubreq_client.get.return_value = reply

        response = self.client.get(
            reverse("ferda_pubreq:api:get", kwargs={"public_request_id": id})
        )

        self.assertEqual(pubreq_client.mock_calls, [call.get(id)])
        self.assertEqual(response.status_code, 200)
        self.assertLog(
            LogEntryType.PUBLIC_REQUEST_GET,
            LoggerResult.SUCCESS,
            input_properties={"public_request_id": id},
            output_references={
                LogReferenceType.DOMAIN: ref_value,
            },
        )

    @patch("ferda_pubreq.api.PUBREQ_INFO", autospec=True)
    def test_get_not_found(self, pubreq_client):
        id: str = "ID 1"
        pubreq_client.get.side_effect = PublicRequestDoesNotExist

        response = self.client.get(
            reverse("ferda_pubreq:api:get", kwargs={"public_request_id": id})
        )

        self.assertEqual(pubreq_client.mock_calls, [call.get(id)])
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.json(), {"detail": "Public request not found"})
        self.assertLog(
            LogEntryType.PUBLIC_REQUEST_GET,
            LoggerResult.FAIL,
            input_properties={"public_request_id": id},
        )

    # --- list

    @patch("ferda_pubreq.api.PUBREQ_INFO", autospec=True)
    def test_list_empty(self, pubreq_client):
        query_param = {"ref_type": "domain", "ref_value": "012-345-6789"}
        pubreq_client.list.return_value = []

        url = reverse("ferda_pubreq:api:list") + "?" + urlencode(query_param)
        response = self.client.get(url)

        self.assertEqual(
            pubreq_client.mock_calls,
            [call.list(query_param["ref_type"], query_param["ref_value"])],
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), [])
        self.assertLog(
            LogEntryType.PUBLIC_REQUEST_LIST,
            LoggerResult.SUCCESS,
            input_properties=query_param,
        )

    @patch("ferda_pubreq.api.PUBREQ_INFO", autospec=True)
    def test_list_full(self, pubreq_client):
        query_param = {"ref_type": "domain", "ref_value": "012-345-6789"}
        reply = [
            make_public_request(id=i + 1, ref_value=query_param["ref_value"]) for i in range(8)
        ]
        pubreq_client.list.return_value = reply

        url = reverse("ferda_pubreq:api:list") + "?" + urlencode(query_param)
        response = self.client.get(url)

        self.assertEqual(
            pubreq_client.mock_calls,
            [call.list(query_param["ref_type"], query_param["ref_value"])],
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual([PublicRequest(**i) for i in response.json()], reply)
        self.assertLog(
            LogEntryType.PUBLIC_REQUEST_LIST,
            LoggerResult.SUCCESS,
            input_properties=query_param,
            output_properties={"items_count": "8"},
            output_references={
                LogReferenceType(query_param["ref_type"]): query_param["ref_value"]
            },
        )

    @patch("ferda_pubreq.api.PUBREQ_INFO", autospec=True)
    def test_list_unknown_registry_object(self, pubreq_client):
        # unknown registry object type raises validation error in function
        query_param = {"type": "???", "value": "012-345-6789"}
        pubreq_client.list.return_value = ""

        url = reverse("ferda_pubreq:api:list") + "?" + urlencode(query_param)
        response = self.client.get(url)

        self.assertEqual(response.status_code, 422)

    @patch("ferda_pubreq.api.PUBREQ_INFO", autospec=True)
    def test_list_missing_param(self, pubreq_client):
        query_param = {"type": "domain"}
        pubreq_client.list.return_value = []

        url = reverse("ferda_pubreq:api:list") + "?" + urlencode(query_param)
        response = self.client.get(url)

        self.assertEqual(response.status_code, 422)


@tag("api")
@override_settings(ROOT_URLCONF="ferda_pubreq.tests.urls")
class PutTest(LoggerMixin, TestCase):
    """Test cases with HTTP PUT method (processing)."""

    def setUp(self):
        super().setUp()
        user = User.objects.create_superuser("admin")
        self.client.force_login(user)

    # --- resolve

    @patch("ferda_pubreq.api.PUBREQ_PROCESSING", autospec=True)
    def test_resolve_ok(self, pubreq_client):
        id = PublicRequestId("ID 1")
        pubreq_client.resolve.return_value = None

        response = self.client.put(
            reverse("ferda_pubreq:api:resolve", kwargs={"public_request_id": id})
        )

        self.assertEqual(pubreq_client.mock_calls, [call.resolve(id, sentinel.log_entry_id)])
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"")
        self.assertLog(
            LogEntryType.PUBLIC_REQUEST_RESOLVE,
            LoggerResult.SUCCESS,
            input_properties={"public_request_id": id},
            output_references={LogReferenceType.PUBLIC_REQUEST: id},
        )

    @patch("ferda_pubreq.api.PUBREQ_PROCESSING", autospec=True)
    def test_resolve_not_found(self, pubreq_client):
        id = PublicRequestId("ID 1")
        pubreq_client.resolve.side_effect = PublicRequestDoesNotExist

        response = self.client.put(
            reverse("ferda_pubreq:api:resolve", kwargs={"public_request_id": id})
        )

        self.assertEqual(pubreq_client.mock_calls, [call.resolve(id, sentinel.log_entry_id)])
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.json(), {"detail": "Public request not found"})
        self.assertLog(
            LogEntryType.PUBLIC_REQUEST_RESOLVE,
            LoggerResult.FAIL,
            input_properties={"public_request_id": id},
        )

    @patch("ferda_pubreq.api.PUBREQ_PROCESSING", autospec=True)
    def test_resolve_failed(self, pubreq_client):
        id = PublicRequestId("ID 1")
        pubreq_client.resolve.side_effect = CannotBeProcessed

        response = self.client.put(
            reverse("ferda_pubreq:api:resolve", kwargs={"public_request_id": id})
        )

        self.assertEqual(pubreq_client.mock_calls, [call.resolve(id, sentinel.log_entry_id)])
        self.assertEqual(response.status_code, 409)
        self.assertEqual(response.json(), {"detail": "Public request cannot be processed"})
        self.assertLog(
            LogEntryType.PUBLIC_REQUEST_RESOLVE,
            LoggerResult.FAIL,
            input_properties={"public_request_id": id},
            output_references={LogReferenceType.PUBLIC_REQUEST: id},
        )

    # --- invalidate

    @patch("ferda_pubreq.api.PUBREQ_PROCESSING", autospec=True)
    def test_invalidate_ok(self, pubreq_client):
        id = PublicRequestId("ID 1")
        pubreq_client.invalidate.return_value = None

        response = self.client.put(
            reverse("ferda_pubreq:api:invalidate", kwargs={"public_request_id": id})
        )

        self.assertEqual(pubreq_client.mock_calls, [call.invalidate(id, sentinel.log_entry_id)])
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"")
        self.assertLog(
            LogEntryType.PUBLIC_REQUEST_INVALIDATE,
            LoggerResult.SUCCESS,
            input_properties={"public_request_id": id},
            output_references={LogReferenceType.PUBLIC_REQUEST: id},
        )

    @patch("ferda_pubreq.api.PUBREQ_PROCESSING", autospec=True)
    def test_invalidate_not_found(self, pubreq_client):
        id = PublicRequestId("ID 1")
        pubreq_client.invalidate.side_effect = PublicRequestDoesNotExist

        response = self.client.put(
            reverse("ferda_pubreq:api:invalidate", kwargs={"public_request_id": id})
        )

        self.assertEqual(pubreq_client.mock_calls, [call.invalidate(id, sentinel.log_entry_id)])
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.json(), {"detail": "Public request not found"})
        self.assertLog(
            LogEntryType.PUBLIC_REQUEST_INVALIDATE,
            LoggerResult.FAIL,
            input_properties={"public_request_id": id},
        )

    @patch("ferda_pubreq.api.PUBREQ_PROCESSING", autospec=True)
    def test_invalidate_failed(self, pubreq_client):
        id = PublicRequestId("ID 1")
        pubreq_client.invalidate.side_effect = CannotBeProcessed

        response = self.client.put(
            reverse("ferda_pubreq:api:invalidate", kwargs={"public_request_id": id})
        )

        self.assertEqual(pubreq_client.mock_calls, [call.invalidate(id, sentinel.log_entry_id)])
        self.assertEqual(response.status_code, 409)
        self.assertEqual(response.json(), {"detail": "Public request cannot be processed"})
        self.assertLog(
            LogEntryType.PUBLIC_REQUEST_INVALIDATE,
            LoggerResult.FAIL,
            input_properties={"public_request_id": id},
            output_references={LogReferenceType.PUBLIC_REQUEST: id},
        )
